# -*- coding: utf-8 -*-
from .jsregexp import Exec
import re
DIGS = set('0123456789')
WHITE = u"\u0009\u000A\u000B\u000C\u000D\u0020\u00A0\u1680\u180E\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200A\u2028\u2029\u202F\u205F\u3000\uFEFF"


def replacement_template(rep, source, span, npar):
    """Takes the replacement template and some info about the match and returns filled template
       """
    n = 0
    res = ''
    while n < len(rep) - 1:
        char = rep[n]
        if char == '$':
            if rep[n + 1] == '$':
                res += '$'
                n += 2
                continue
            elif rep[n + 1] == '&':
                # replace with matched string
                res += source[span[0]:span[1]]
                n += 2
                continue
            elif rep[n + 1] == '`':
                # replace with string that is BEFORE match
                res += source[:span[0]]
                n += 2
                continue
            elif rep[n + 1] == '\'':
                # replace with string that is AFTER match
                res += source[span[1]:]
                n += 2
                continue
            elif rep[n + 1] in DIGS:
                dig = rep[n + 1]
                if n + 2 < len(rep) and rep[n + 2] in DIGS:
                    dig += rep[n + 2]
                num = int(dig)
                # we will not do any replacements if we dont have this npar or dig is 0
                if not num or num > len(npar):
                    res += '$' + dig
                else:
                    # None - undefined has to be replaced with ''
                    res += npar[num - 1] if npar[num - 1] else ''
                n += 1 + len(dig)
                continue
        res += char
        n += 1
    if n < len(rep):
        res += rep[-1]
    return res


###################################################


class StringPrototype:
    def toString():
        if this.Class != 'String':
            raise this.MakeError('TypeError',
                                 'String.prototype.toString is not generic')
        return this.value

    def valueOf():
        if this.Class != 'String':
            raise this.MakeError('TypeError',
                                 'String.prototype.valueOf is not generic')
        return this.value

    def charAt(pos):
        this.cok()
        pos = pos.to_int()
        s = this.to_string()
        if 0 <= pos < len(s.value):
            char = s.value[pos]
            if char not in s.CHAR_BANK:
                s.Js(char)  # add char to char bank
            return s.CHAR_BANK[char]
        return s.CHAR_BANK['']

    def charCodeAt(pos):
        this.cok()
        pos = pos.to_int()
        s = this.to_string()
        if 0 <= pos < len(s.value):
            return s.Js(ord(s.value[pos]))
        return s.NaN

    def concat():
        this.cok()
        s = this.to_string()
        res = s.value
        for e in arguments.to_list():
            res += e.to_string().value
        return res

    def indexOf(searchString, position):
        this.cok()
        s = this.to_string().value
        search = searchString.to_string().value
        pos = position.to_int()
        return this.Js(s.find(search, min(max(pos, 0), len(s))))

    def lastIndexOf(searchString, position):
        this.cok()
        s = this.to_string().value
        search = searchString.to_string().value
        pos = position.to_number()
        pos = 10**15 if pos.is_nan() else pos.to_int()
        return s.rfind(search, 0, min(max(pos, 0) + 1, len(s)))

    def localeCompare(that):
        this.cok()
        s = this.to_string()
        that = that.to_string()
        if s < that:
            return this.Js(-1)
        elif s > that:
            return this.Js(1)
        return this.Js(0)

    def match(regexp):
        this.cok()
        s = this.to_string()
        r = this.RegExp(regexp) if regexp.Class != 'RegExp' else regexp
        if not r.glob:
            return Exec(r, s)
        r.put('lastIndex', this.Js(0))
        found = []
        previous_last_index = 0
        last_match = True
        while last_match:
            result = Exec(r, s)
            if result.is_null():
                last_match = False
            else:
                this_index = r.get('lastIndex').value
                if this_index == previous_last_index:
                    r.put('lastIndex', this.Js(this_index + 1))
                    previous_last_index += 1
                else:
                    previous_last_index = this_index
                matchStr = result.get('0')
                found.append(matchStr)
        if not found:
            return this.null
        return found

    def replace(searchValue, replaceValue):
        # VERY COMPLICATED. to check again.
        this.cok()
        string = this.to_string()
        s = string.value
        res = ''
        if not replaceValue.is_callable():
            replaceValue = replaceValue.to_string().value
            func = False
        else:
            func = True
        # Replace all  ( global )
        if searchValue.Class == 'RegExp' and searchValue.glob:
            last = 0
            for e in re.finditer(searchValue.pat, s):
                res += s[last:e.span()[0]]
                if func:
                    # prepare arguments for custom func (replaceValue)
                    args = (e.group(), ) + e.groups() + (e.span()[1], string)
                    # convert all types to JS
                    args = map(this.Js, args)
                    res += replaceValue(*args).to_string().value
                else:
                    res += replacement_template(replaceValue, s, e.span(),
                                                e.groups())
                last = e.span()[1]
            res += s[last:]
            return this.Js(res)
        elif searchValue.Class == 'RegExp':
            e = re.search(searchValue.pat, s)
            if e is None:
                return string
            span = e.span()
            pars = e.groups()
            match = e.group()
        else:
            match = searchValue.to_string().value
            ind = s.find(match)
            if ind == -1:
                return string
            span = ind, ind + len(match)
            pars = ()
        res = s[:span[0]]
        if func:
            args = (match, ) + pars + (span[1], string)
            # convert all types to JS
            this_ = this
            args = tuple([this_.Js(x) for x in args])
            res += replaceValue(*args).to_string().value
        else:
            res += replacement_template(replaceValue, s, span, pars)
        res += s[span[1]:]
        return res

    def search(regexp):
        this.cok()
        string = this.to_string()
        if regexp.Class == 'RegExp':
            rx = regexp
        else:
            rx = this.RegExp(regexp)
        res = re.search(rx.pat, string.value)
        if res is not None:
            return this.Js(res.span()[0])
        return -1

    def slice(start, end):
        this.cok()
        s = this.to_string()
        start = start.to_int()
        length = len(s.value)
        end = length if end.is_undefined() else end.to_int()
        #From = max(length+start, 0) if start<0 else min(length, start)
        #To = max(length+end, 0) if end<0 else min(length, end)
        return s.value[start:end]

    def split(separator, limit):
        # its a bit different that re.split!
        this.cok()
        S = this.to_string()
        s = S.value
        lim = 2**32 - 1 if limit.is_undefined() else limit.to_uint32()
        if not lim:
            return []
        if separator.is_undefined():
            return [s]
        len_s = len(s)
        res = []
        R = separator if separator.Class == 'RegExp' else separator.to_string()
        if not len_s:
            if SplitMatch(s, 0, R) is None:
                return [S]
            return []
        p = q = 0
        while q != len_s:
            e, cap = SplitMatch(s, q, R)
            if e is None or e == p:
                q += 1
                continue
            res.append(s[p:q])
            p = q = e
            if len(res) == lim:
                return res
            for element in cap:
                res.append(this.Js(element))
                if len(res) == lim:
                    return res
        res.append(s[p:])
        return res

    def substring(start, end):
        this.cok()
        s = this.to_string().value
        start = start.to_int()
        length = len(s)
        end = length if end.is_undefined() else end.to_int()
        fstart = min(max(start, 0), length)
        fend = min(max(end, 0), length)
        return this.Js(s[min(fstart, fend):max(fstart, fend)])

    def substr(start, length):
        #I hate this function and its description in specification
        r1 = this.to_string().value
        r2 = start.to_int()
        r3 = 10**20 if length.is_undefined() else length.to_int()
        r4 = len(r1)
        r5 = r2 if r2 >= 0 else max(0, r2 + r4)
        r6 = min(max(r3, 0), r4 - r5)
        if r6 <= 0:
            return ''
        return r1[r5:r5 + r6]

    def toLowerCase():
        this.cok()
        return this.Js(this.to_string().value.lower())

    def toLocaleLowerCase():
        this.cok()
        return this.Js(this.to_string().value.lower())

    def toUpperCase():
        this.cok()
        return this.Js(this.to_string().value.upper())

    def toLocaleUpperCase():
        this.cok()
        return this.Js(this.to_string().value.upper())

    def trim():
        this.cok()
        return this.Js(this.to_string().value.strip(WHITE))


def SplitMatch(s, q, R):
    # s is Py String to match, q is the py int match start and R is Js RegExp or String.
    if R.Class == 'RegExp':
        res = R.match(s, q)
        return (None, ()) if res is None else (res.span()[1], res.groups())
    # R is just a string
    if s[q:].startswith(R.value):
        return q + len(R.value), ()
    return None, ()



from __future__ import print_function

from timeit import timeit
from collections import namedtuple
from array import array
try:
    #python 2 code
    from itertools import izip as zip
except ImportError:
    pass

from collections import deque


class Y(object):
    UUU = 88

    def __init__(self, x):
        self.x = x

    def s(self, x):
        return self.x + 1


class X(Y):
    A = 10
    B = 2
    C = 4
    D = 9

    def __init__(self, x):
        self.x = x
        self.stack = []
        self.par = super(X, self)

    def s(self, x):
        pass

    def __add__(self, other):
        return self.x + other.x

    def another(self):
        return Y.s(self, 1)

    def yet_another(self):
        return self.par.s(1)


def add(a, b):
    return a.x + b.x


t = []

Type = None
try:
    print(timeit(
        """

t.append(4)
t.pop()



""",
        "from __main__ import X,Y,namedtuple,array,t,add,Type, zip",
        number=1000000))
except:
    raise



from ..base import *
import math
import random

Math = PyJsObject(prototype=ObjectPrototype)
Math.Class = 'Math'

CONSTANTS = {
    'E': 2.7182818284590452354,
    'LN10': 2.302585092994046,
    'LN2': 0.6931471805599453,
    'LOG2E': 1.4426950408889634,
    'LOG10E': 0.4342944819032518,
    'PI': 3.1415926535897932,
    'SQRT1_2': 0.7071067811865476,
    'SQRT2': 1.4142135623730951
}

for constant, value in CONSTANTS.items():
    Math.define_own_property(
        constant, {
            'value': Js(value),
            'writable': False,
            'enumerable': False,
            'configurable': False
        })


class MathFunctions:
    def abs(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        return abs(a)

    def acos(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        try:
            return math.acos(a)
        except:
            return NaN

    def asin(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        try:
            return math.asin(a)
        except:
            return NaN

    def atan(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        return math.atan(a)

    def atan2(y, x):
        a = x.to_number().value
        b = y.to_number().value
        if a != a or b != b:  # it must be a nan
            return NaN
        return math.atan2(b, a)

    def ceil(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        return math.ceil(a)

    def floor(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        return math.floor(a)

    def round(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        return round(a)

    def sin(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        return math.sin(a)

    def cos(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        return math.cos(a)

    def tan(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        return math.tan(a)

    def log(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        try:
            return math.log(a)
        except:
            return NaN

    def exp(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        return math.exp(a)

    def pow(x, y):
        a = x.to_number().value
        b = y.to_number().value
        if a != a or b != b:  # it must be a nan
            return NaN
        try:
            return a**b
        except:
            return NaN

    def sqrt(x):
        a = x.to_number().value
        if a != a:  # it must be a nan
            return NaN
        try:
            return a**0.5
        except:
            return NaN

    def min():
        if not len(arguments):
            return Infinity
        lis = tuple(e.to_number().value for e in arguments.to_list())
        if any(e != e for e in lis):  # we dont want NaNs
            return NaN
        return min(*lis)

    def max():
        if not len(arguments):
            return -Infinity
        lis = tuple(e.to_number().value for e in arguments.to_list())
        if any(e != e for e in lis):  # we dont want NaNs
            return NaN
        return max(*lis)

    def random():
        return random.random()


fill_prototype(Math, MathFunctions, default_attrs)



from __future__ import unicode_literals

from ..conversions import *
from ..func_utils import *

import math
import random

CONSTANTS = {
    'E': 2.7182818284590452354,
    'LN10': 2.302585092994046,
    'LN2': 0.6931471805599453,
    'LOG2E': 1.4426950408889634,
    'LOG10E': 0.4342944819032518,
    'PI': 3.1415926535897932,
    'SQRT1_2': 0.7071067811865476,
    'SQRT2': 1.4142135623730951
}
def is_infinity(x):
    return x - 1e10 == x

class MathFunctions:
    def abs(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if a != a:  # it must be a nan
            return NaN
        return abs(a)

    def acos(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if a != a:  # it must be a nan
            return NaN
        try:
            return math.acos(a)
        except:
            return NaN

    def asin(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if a != a:  # it must be a nan
            return NaN
        try:
            return math.asin(a)
        except:
            return NaN

    def atan(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if a != a:  # it must be a nan
            return NaN
        return math.atan(a)

    def atan2(this, args):
        x = get_arg(args, 0)
        y = get_arg(args, 1)
        a = to_number(x)
        b = to_number(y)
        if a != a or b != b:  # it must be a nan
            return NaN
        return math.atan2(a, b)

    def ceil(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if not is_finite(x):
            return x
        return float(math.ceil(a))

    def floor(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if not is_finite(x):
            return x
        return float(math.floor(a))

    def round(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if not is_finite(x):
            return x
        return float(round(a))

    def sin(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if not is_finite(a):  # it must be a nan
            return NaN
        return math.sin(a)

    def cos(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if not is_finite(a):  # it must be a nan
            return NaN
        return math.cos(a)

    def tan(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if not is_finite(a):  # it must be a nan
            return NaN
        return math.tan(a)

    def log(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if a != a:  # it must be a nan
            return NaN
        try:
            return math.log(a)
        except:
            return NaN

    def exp(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if a != a:  # it must be a nan
            return NaN
        return math.exp(a)

    def pow(this, args):
        x = get_arg(args, 0)
        y = get_arg(args, 1)
        a = to_number(x)
        b = to_number(y)
        if a != a or b != b:  # it must be a nan
            return NaN
        try:
            return a**b
        except:
            return NaN

    def sqrt(this, args):
        x = get_arg(args, 0)
        a = to_number(x)
        if a != a:  # it must be a nan
            return NaN
        try:
            return a**0.5
        except:
            return NaN

    def min(this, args):
        if len(args) == 0:
            return Infinity
        return min(map(to_number, tuple(args)))

    def max(this, args):
        if len(args) == 0:
            return -Infinity
        return max(map(to_number, tuple(args)))

    def random(this, args):
        return random.random()



from .code import Code
from .simplex import MakeError
from .opcodes import *
from .operations import *
from .trans_utils import *

SPECIAL_IDENTIFIERS = {'true', 'false', 'this'}


class ByteCodeGenerator:
    def __init__(self, exe):
        self.exe = exe

        self.declared_continue_labels = {}
        self.declared_break_labels = {}

        self.implicit_breaks = []
        self.implicit_continues = []

        self.declared_vars = []

        self.function_declaration_tape = []

        self.states = []

    def record_state(self):
        self.states.append(
            (self.declared_continue_labels, self.declared_break_labels,
             self.implicit_breaks, self.implicit_continues, self.declared_vars,
             self.function_declaration_tape))
        self.declared_continue_labels, self.declared_break_labels, \
        self.implicit_breaks, self.implicit_continues, \
        self.declared_vars, self.function_declaration_tape = {}, {}, [], [], [], []

    def restore_state(self):
        self.declared_continue_labels, self.declared_break_labels, \
        self.implicit_breaks, self.implicit_continues, \
        self.declared_vars, self.function_declaration_tape = self.states.pop()

    def ArrayExpression(self, elements, **kwargs):
        for e in elements:
            if e is None:
                self.emit('LOAD_NONE')
            else:
                self.emit(e)
        self.emit('LOAD_ARRAY', len(elements))

    def AssignmentExpression(self, operator, left, right, **kwargs):
        operator = operator[:-1]
        if left['type'] == 'MemberExpression':
            self.emit(left['object'])
            if left['computed']:
                self.emit(left['property'])
                self.emit(right)
                if operator:
                    self.emit('STORE_MEMBER_OP', operator)
                else:
                    self.emit('STORE_MEMBER')
            else:
                self.emit(right)
                if operator:
                    self.emit('STORE_MEMBER_DOT_OP', left['property']['name'],
                              operator)
                else:
                    self.emit('STORE_MEMBER_DOT', left['property']['name'])
        elif left['type'] == 'Identifier':
            if left['name'] in SPECIAL_IDENTIFIERS:
                raise MakeError('SyntaxError',
                                'Invalid left-hand side in assignment')
            self.emit(right)
            if operator:
                self.emit('STORE_OP', left['name'], operator)
            else:
                self.emit('STORE', left['name'])
        else:
            raise MakeError('SyntaxError',
                            'Invalid left-hand side in assignment')

    def BinaryExpression(self, operator, left, right, **kwargs):
        self.emit(left)
        self.emit(right)
        self.emit('BINARY_OP', operator)

    def BlockStatement(self, body, **kwargs):
        self._emit_statement_list(body)

    def BreakStatement(self, label, **kwargs):
        if label is None:
            self.emit('JUMP', self.implicit_breaks[-1])
        else:
            label = label.get('name')
            if label not in self.declared_break_labels:
                raise MakeError('SyntaxError',
                                'Undefined label \'%s\'' % label)
            else:
                self.emit('JUMP', self.declared_break_labels[label])

    def CallExpression(self, callee, arguments, **kwargs):
        if callee['type'] == 'MemberExpression':
            self.emit(callee['object'])
            if callee['computed']:
                self.emit(callee['property'])
                if arguments:
                    for e in arguments:
                        self.emit(e)
                    self.emit('LOAD_N_TUPLE', len(arguments))
                    self.emit('CALL_METHOD')
                else:
                    self.emit('CALL_METHOD_NO_ARGS')
            else:
                prop_name = to_key(callee['property'])
                if arguments:
                    for e in arguments:
                        self.emit(e)
                    self.emit('LOAD_N_TUPLE', len(arguments))
                    self.emit('CALL_METHOD_DOT', prop_name)
                else:
                    self.emit('CALL_METHOD_DOT_NO_ARGS', prop_name)
        else:
            self.emit(callee)
            if arguments:
                for e in arguments:
                    self.emit(e)
                self.emit('LOAD_N_TUPLE', len(arguments))
                self.emit('CALL')
            else:
                self.emit('CALL_NO_ARGS')

    def ClassBody(self, body, **kwargs):
        raise NotImplementedError('Not available in ECMA 5.1')

    def ClassDeclaration(self, id, superClass, body, **kwargs):
        raise NotImplementedError('Not available in ECMA 5.1')

    def ClassExpression(self, id, superClass, body, **kwargs):
        raise NotImplementedError('Classes not available in ECMA 5.1')

    def ConditionalExpression(self, test, consequent, alternate, **kwargs):
        alt = self.exe.get_new_label()
        end = self.exe.get_new_label()
        # ?
        self.emit(test)
        self.emit('JUMP_IF_FALSE', alt)
        # first val
        self.emit(consequent)
        self.emit('JUMP', end)
        # second val
        self.emit('LABEL', alt)
        self.emit(alternate)
        # end of ?: statement
        self.emit('LABEL', end)

    def ContinueStatement(self, label, **kwargs):
        if label is None:
            self.emit('JUMP', self.implicit_continues[-1])
        else:
            label = label.get('name')
            if label not in self.declared_continue_labels:
                raise MakeError('SyntaxError',
                                'Undefined label \'%s\'' % label)
            else:
                self.emit('JUMP', self.declared_continue_labels[label])

    def DebuggerStatement(self, **kwargs):
        self.EmptyStatement(**kwargs)

    def DoWhileStatement(self, body, test, **kwargs):
        continue_label = self.exe.get_new_label()
        break_label = self.exe.get_new_label()
        initial_do = self.exe.get_new_label()

        self.emit('JUMP', initial_do)
        self.emit('LABEL', continue_label)
        self.emit(test)
        self.emit('JUMP_IF_FALSE', break_label)
        self.emit('LABEL', initial_do)

        # translate the body, remember to add and afterwards remove implicit break/continue labels

        self.implicit_continues.append(continue_label)
        self.implicit_breaks.append(break_label)
        self.emit(body)
        self.implicit_continues.pop()
        self.implicit_breaks.pop()

        self.emit('JUMP', continue_label)  # loop back
        self.emit('LABEL', break_label)

    def EmptyStatement(self, **kwargs):
        # do nothing
        pass

    def ExpressionStatement(self, expression, **kwargs):
        # change the final stack value
        # pop the previous value and execute expression
        self.emit('POP')
        self.emit(expression)

    def ForStatement(self, init, test, update, body, **kwargs):
        continue_label = self.exe.get_new_label()
        break_label = self.exe.get_new_label()
        first_start = self.exe.get_new_label()

        if init is not None:
            self.emit(init)
            if init['type'] != 'VariableDeclaration':
                self.emit('POP')

        # skip first update and go straight to test
        self.emit('JUMP', first_start)

        self.emit('LABEL', continue_label)
        if update:
            self.emit(update)
            self.emit('POP')
        self.emit('LABEL', first_start)
        if test:
            self.emit(test)
            self.emit('JUMP_IF_FALSE', break_label)

        # translate the body, remember to add and afterwards to remove implicit break/continue labels

        self.implicit_continues.append(continue_label)
        self.implicit_breaks.append(break_label)
        self.emit(body)
        self.implicit_continues.pop()
        self.implicit_breaks.pop()

        self.emit('JUMP', continue_label)  # loop back
        self.emit('LABEL', break_label)

    def ForInStatement(self, left, right, body, **kwargs):
        # prepare the needed labels
        body_start_label = self.exe.get_new_label()
        continue_label = self.exe.get_new_label()
        break_label = self.exe.get_new_label()

        # prepare the name
        if left['type'] == 'VariableDeclaration':
            if len(left['declarations']) != 1:
                raise MakeError(
                    'SyntaxError',
                    ' Invalid left-hand side in for-in loop: Must have a single binding.'
                )
            self.emit(left)
            name = left['declarations'][0]['id']['name']
        elif left['type'] == 'Identifier':
            name = left['name']
        else:
            raise MakeError('SyntaxError',
                            'Invalid left-hand side in for-loop')

        # prepare the iterable
        self.emit(right)

        # emit ForIn Opcode
        self.emit('FOR_IN', name, body_start_label, continue_label,
                  break_label)

        # a special continue position
        self.emit('LABEL', continue_label)
        self.emit('NOP')

        self.emit('LABEL', body_start_label)
        self.implicit_continues.append(continue_label)
        self.implicit_breaks.append(break_label)
        self.emit('LOAD_UNDEFINED')
        self.emit(body)
        self.implicit_continues.pop()
        self.implicit_breaks.pop()
        self.emit('NOP')
        self.emit('LABEL', break_label)
        self.emit('NOP')

    def FunctionDeclaration(self, id, params, defaults, body, **kwargs):
        if defaults:
            raise NotImplementedError('Defaults not available in ECMA 5.1')

        # compile function
        self.record_state(
        )  # cleans translator state and appends it to the stack so that it can be later restored
        function_start = self.exe.get_new_label()
        function_declarations = self.exe.get_new_label()
        declarations_done = self.exe.get_new_label(
        )  # put jump to this place at the and of function tape!
        function_end = self.exe.get_new_label()

        # skip the function if encountered externally
        self.emit('JUMP', function_end)

        self.emit('LABEL', function_start)
        # call is made with empty stack so load undefined to fill it
        self.emit('LOAD_UNDEFINED')
        # declare all functions
        self.emit('JUMP', function_declarations)
        self.emit('LABEL', declarations_done)
        self.function_declaration_tape.append(LABEL(function_declarations))

        self.emit(body)
        self.ReturnStatement(None)

        self.function_declaration_tape.append(JUMP(declarations_done))
        self.exe.tape.extend(self.function_declaration_tape)

        self.emit('LABEL', function_end)
        declared_vars = self.declared_vars
        self.restore_state()

        # create function object and append to stack
        name = id.get('name')
        assert name is not None
        self.declared_vars.append(name)
        self.function_declaration_tape.append(
            LOAD_FUNCTION(function_start, tuple(p['name'] for p in params),
                          name, True, tuple(declared_vars)))
        self.function_declaration_tape.append(STORE(name))
        self.function_declaration_tape.append(POP())

    def FunctionExpression(self, id, params, defaults, body, **kwargs):
        if defaults:
            raise NotImplementedError('Defaults not available in ECMA 5.1')

        # compile function
        self.record_state(
        )  # cleans translator state and appends it to the stack so that it can be later restored
        function_start = self.exe.get_new_label()
        function_declarations = self.exe.get_new_label()
        declarations_done = self.exe.get_new_label(
        )  # put jump to this place at the and of function tape!
        function_end = self.exe.get_new_label()

        # skip the function if encountered externally
        self.emit('JUMP', function_end)

        self.emit('LABEL', function_start)
        # call is made with empty stack so load undefined to fill it
        self.emit('LOAD_UNDEFINED')
        # declare all functions
        self.emit('JUMP', function_declarations)
        self.emit('LABEL', declarations_done)
        self.function_declaration_tape.append(LABEL(function_declarations))

        self.emit(body)
        self.ReturnStatement(None)

        self.function_declaration_tape.append(JUMP(declarations_done))
        self.exe.tape.extend(self.function_declaration_tape)

        self.emit('LABEL', function_end)
        declared_vars = self.declared_vars
        self.restore_state()

        # create function object and append to stack
        name = id.get('name') if id else None
        self.emit('LOAD_FUNCTION', function_start,
                  tuple(p['name'] for p in params), name, False,
                  tuple(declared_vars))

    def Identifier(self, name, **kwargs):
        if name == 'true':
            self.emit('LOAD_BOOLEAN', 1)
        elif name == 'false':
            self.emit('LOAD_BOOLEAN', 0)
        elif name == 'undefined':
            self.emit('LOAD_UNDEFINED')
        else:
            self.emit('LOAD', name)

    def IfStatement(self, test, consequent, alternate, **kwargs):
        alt = self.exe.get_new_label()
        end = self.exe.get_new_label()
        # if
        self.emit(test)
        self.emit('JUMP_IF_FALSE', alt)
        # consequent
        self.emit(consequent)
        self.emit('JUMP', end)
        # alternate
        self.emit('LABEL', alt)
        if alternate is not None:
            self.emit(alternate)
        # end of if statement
        self.emit('LABEL', end)

    def LabeledStatement(self, label, body, **kwargs):
        label = label['name']
        if body['type'] in ('WhileStatement', 'DoWhileStatement',
                            'ForStatement', 'ForInStatement'):
            # Continue label available... Simply take labels defined by the loop.
            # It is important that they request continue label first
            self.declared_continue_labels[label] = self.exe._label_count + 1
            self.declared_break_labels[label] = self.exe._label_count + 2
            self.emit(body)
            del self.declared_break_labels[label]
            del self.declared_continue_labels[label]
        else:
            # only break label available
            lbl = self.exe.get_new_label()
            self.declared_break_labels[label] = lbl
            self.emit(body)
            self.emit('LABEL', lbl)
            del self.declared_break_labels[label]

    def Literal(self, value, **kwargs):
        if value is None:
            self.emit('LOAD_NULL')
        elif isinstance(value, bool):
            self.emit('LOAD_BOOLEAN', int(value))
        elif isinstance(value, basestring):
            self.emit('LOAD_STRING', unicode(value))
        elif isinstance(value, (float, int, long)):
            self.emit('LOAD_NUMBER', float(value))
        elif isinstance(value, tuple):
            self.emit('LOAD_REGEXP', *value)
        else:
            raise RuntimeError('Unsupported literal')

    def LogicalExpression(self, left, right, operator, **kwargs):
        end = self.exe.get_new_label()
        if operator == '&&':
            # AND
            self.emit(left)
            self.emit('JUMP_IF_FALSE_WITHOUT_POP', end)
            self.emit('POP')
            self.emit(right)
            self.emit('LABEL', end)
        elif operator == '||':
            # OR
            self.emit(left)
            self.emit('JUMP_IF_TRUE_WITHOUT_POP', end)
            self.emit('POP')
            self.emit(right)
            self.emit('LABEL', end)
        else:
            raise RuntimeError("Unknown logical expression: %s" % operator)

    def MemberExpression(self, computed, object, property, **kwargs):
        if computed:
            self.emit(object)
            self.emit(property)
            self.emit('LOAD_MEMBER')
        else:
            self.emit(object)
            self.emit('LOAD_MEMBER_DOT', property['name'])

    def NewExpression(self, callee, arguments, **kwargs):
        self.emit(callee)
        if arguments:
            n = len(arguments)
            for e in arguments:
                self.emit(e)
            self.emit('LOAD_N_TUPLE', n)
            self.emit('NEW')
        else:
            self.emit('NEW_NO_ARGS')

    def ObjectExpression(self, properties, **kwargs):
        data = []
        for prop in properties:
            self.emit(prop['value'])
            if prop['computed']:
                raise NotImplementedError(
                    'ECMA 5.1 does not support computed object properties!')
            data.append((to_key(prop['key']), prop['kind'][0]))
        self.emit('LOAD_OBJECT', tuple(data))

    def Program(self, body, **kwargs):
        old_tape_len = len(self.exe.tape)
        self.emit('LOAD_UNDEFINED')
        self.emit(body)
        # add function tape !
        self.exe.tape = self.exe.tape[:old_tape_len] + self.function_declaration_tape + self.exe.tape[old_tape_len:]

    def Pyimport(self, imp, **kwargs):
        raise NotImplementedError(
            'Not available for bytecode interpreter yet, use the Js2Py translator.'
        )

    def Property(self, kind, key, computed, value, method, shorthand,
                 **kwargs):
        raise NotImplementedError('Not available in ECMA 5.1')

    def RestElement(self, argument, **kwargs):
        raise NotImplementedError('Not available in ECMA 5.1')

    def ReturnStatement(self, argument, **kwargs):
        self.emit('POP')  # pop result of expression statements
        if argument is None:
            self.emit('LOAD_UNDEFINED')
        else:
            self.emit(argument)
        self.emit('RETURN')

    def SequenceExpression(self, expressions, **kwargs):
        for e in expressions:
            self.emit(e)
            self.emit('POP')
        del self.exe.tape[-1]

    def SwitchCase(self, test, consequent, **kwargs):
        raise NotImplementedError('Already implemented in SwitchStatement')

    def SwitchStatement(self, discriminant, cases, **kwargs):
        self.emit(discriminant)
        labels = [self.exe.get_new_label() for case in cases]
        tests = [case['test'] for case in cases]
        consequents = [case['consequent'] for case in cases]
        end_of_switch = self.exe.get_new_label()

        # translate test cases
        for test, label in zip(tests, labels):
            if test is not None:
                self.emit(test)
                self.emit('JUMP_IF_EQ', label)
            else:
                self.emit('POP')
                self.emit('JUMP', label)
        # this will be executed if none of the cases worked
        self.emit('POP')
        self.emit('JUMP', end_of_switch)

        # translate consequents
        self.implicit_breaks.append(end_of_switch)
        for consequent, label in zip(consequents, labels):
            self.emit('LABEL', label)
            self._emit_statement_list(consequent)
        self.implicit_breaks.pop()

        self.emit('LABEL', end_of_switch)

    def ThisExpression(self, **kwargs):
        self.emit('LOAD_THIS')

    def ThrowStatement(self, argument, **kwargs):
        # throw with the empty stack
        self.emit('POP')
        self.emit(argument)
        self.emit('THROW')

    def TryStatement(self, block, handler, finalizer, **kwargs):
        try_label = self.exe.get_new_label()
        catch_label = self.exe.get_new_label()
        finally_label = self.exe.get_new_label()
        end_label = self.exe.get_new_label()

        self.emit('JUMP', end_label)

        # try block
        self.emit('LABEL', try_label)
        self.emit('LOAD_UNDEFINED')
        self.emit(block)
        self.emit(
            'NOP'
        )  # needed to distinguish from break/continue vs some internal jumps

        # catch block
        self.emit('LABEL', catch_label)
        self.emit('LOAD_UNDEFINED')
        if handler:
            self.emit(handler['body'])
        self.emit('NOP')

        # finally block
        self.emit('LABEL', finally_label)
        self.emit('LOAD_UNDEFINED')
        if finalizer:
            self.emit(finalizer)
        self.emit('NOP')

        self.emit('LABEL', end_label)

        # give life to the code
        self.emit('TRY_CATCH_FINALLY', try_label, catch_label,
                  handler['param']['name'] if handler else None, finally_label,
                  bool(finalizer), end_label)

    def UnaryExpression(self, operator, argument, **kwargs):
        if operator == 'typeof' and argument[
                'type'] == 'Identifier':  # todo fix typeof
            self.emit('TYPEOF', argument['name'])
        elif operator == 'delete':
            if argument['type'] == 'MemberExpression':
                self.emit(argument['object'])
                if argument['property']['type'] == 'Identifier':
                    self.emit('LOAD_STRING',
                              unicode(argument['property']['name']))
                else:
                    self.emit(argument['property'])
                self.emit('DELETE_MEMBER')
            elif argument['type'] == 'Identifier':
                self.emit('DELETE', argument['name'])
            else:
                self.emit('LOAD_BOOLEAN', 1)
        elif operator in UNARY_OPERATIONS:
            self.emit(argument)
            self.emit('UNARY_OP', operator)
        else:
            raise MakeError('SyntaxError',
                            'Unknown unary operator %s' % operator)

    def UpdateExpression(self, operator, argument, prefix, **kwargs):
        incr = int(operator == "++")
        post = int(not prefix)
        if argument['type'] == 'MemberExpression':
            if argument['computed']:
                self.emit(argument['object'])
                self.emit(argument['property'])
                self.emit('POSTFIX_MEMBER', post, incr)
            else:
                self.emit(argument['object'])
                name = to_key(argument['property'])
                self.emit('POSTFIX_MEMBER_DOT', post, incr, name)
        elif argument['type'] == 'Identifier':
            name = to_key(argument)
            self.emit('POSTFIX', post, incr, name)
        else:
            raise MakeError('SyntaxError',
                            'Invalid left-hand side in assignment')

    def VariableDeclaration(self, declarations, kind, **kwargs):
        if kind != 'var':
            raise NotImplementedError(
                'Only var variable declaration is supported by ECMA 5.1')
        for d in declarations:
            self.emit(d)

    def LexicalDeclaration(self, declarations, kind, **kwargs):
        raise NotImplementedError('Not supported by ECMA 5.1')

    def VariableDeclarator(self, id, init, **kwargs):
        name = id['name']
        if name in SPECIAL_IDENTIFIERS:
            raise MakeError('Invalid left-hand side in assignment')
        self.declared_vars.append(name)
        if init is not None:
            self.emit(init)
            self.emit('STORE', name)
            self.emit('POP')

    def WhileStatement(self, test, body, **kwargs):
        continue_label = self.exe.get_new_label()
        break_label = self.exe.get_new_label()

        self.emit('LABEL', continue_label)
        self.emit(test)
        self.emit('JUMP_IF_FALSE', break_label)

        # translate the body, remember to add and afterwards remove implicit break/continue labels

        self.implicit_continues.append(continue_label)
        self.implicit_breaks.append(break_label)
        self.emit(body)
        self.implicit_continues.pop()
        self.implicit_breaks.pop()

        self.emit('JUMP', continue_label)  # loop back
        self.emit('LABEL', break_label)

    def WithStatement(self, object, body, **kwargs):
        beg_label = self.exe.get_new_label()
        end_label = self.exe.get_new_label()
        # scope
        self.emit(object)

        # now the body
        self.emit('JUMP', end_label)
        self.emit('LABEL', beg_label)
        self.emit('LOAD_UNDEFINED')
        self.emit(body)
        self.emit('NOP')
        self.emit('LABEL', end_label)

        # with statement implementation
        self.emit('WITH', beg_label, end_label)

    def _emit_statement_list(self, statements):
        for statement in statements:
            self.emit(statement)

    def emit(self, what, *args):
        ''' what can be either name of the op, or node, or a list of statements.'''
        if isinstance(what, basestring):
            return self.exe.emit(what, *args)
        elif isinstance(what, list):
            self._emit_statement_list(what)
        else:
            return getattr(self, what['type'])(**what)


import os, codecs


def path_as_local(path):
    if os.path.isabs(path):
        return path
    # relative to cwd
    return os.path.join(os.getcwd(), path)


def get_file_contents(path_or_file):
    if hasattr(path_or_file, 'read'):
        js = path_or_file.read()
    else:
        with codecs.open(path_as_local(path_or_file), "r", "utf-8") as f:
            js = f.read()
    return js


def main():
    from space import Space
    import fill_space

    from pyjsparser import parse
    import json
    a = ByteCodeGenerator(Code())

    s = Space()
    fill_space.fill_space(s, a)

    a.exe.space = s
    s.exe = a.exe
    con = get_file_contents('internals/esprima.js')
    d = parse(con + (
        ''';JSON.stringify(exports.parse(%s), 4, 4)''' % json.dumps(con)))
    # d = parse('''
    # function x(n) {
    #     log(n)
    #     return x(n+1)
    # }
    # x(0)
    # ''')

    # var v = 333333;
    # while (v) {
    #     v--
    #
    # }
    a.emit(d)
    print(a.declared_vars)
    print(a.exe.tape)
    print(len(a.exe.tape))

    a.exe.compile()

    def log(this, args):
        print(args[0])
        return 999

    print(a.exe.run(a.exe.space.GlobalObj))


if __name__ == '__main__':
    main()



# The MIT License
#
# Copyright 2014, 2015 Piotr Dabkowski
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the 'Software'),
# to deal in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so, subject
# to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
#  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
""" This module allows you to translate and execute Javascript in pure python.
    Basically its implementation of ECMAScript 5.1 in pure python.

    Use eval_js method to execute javascript code and get resulting python object (builtin if possible).

    EXAMPLE:
    >>> import js2py
    >>> add = js2py.eval_js('function add(a, b) {return a + b}')
    >>> add(1, 2) + 3
    6
    >>> add('1', 2, 3)
    u'12'
    >>> add.constructor
    function Function() { [python code] }


    Or use EvalJs to execute many javascript code fragments under same context - you would be able to get any
    variable from the context!

    >>> js = js2py.EvalJs()
    >>> js.execute('var a = 10; function f(x) {return x*x};')
    >>> js.f(9)
    81
    >>> js.a
    10

    Also you can use its console method to play with interactive javascript console.


    Use parse_js to parse (syntax tree is just like in esprima.js) and translate_js to trasnlate JavaScript.

    Finally, you can use pyimport statement from inside JS code to import and use python libraries.

    >>> js2py.eval_js('pyimport urllib; urllib.urlopen("https://www.google.com")')

    NOTE: This module is still not fully finished:

          Date and JSON builtin objects are not implemented
          Array prototype is not fully finished (will be soon)

    Other than that everything should work fine.

"""

__author__ = 'Piotr Dabkowski'
__all__ = [
    'EvalJs', 'translate_js', 'import_js', 'eval_js', 'parse_js',
    'translate_file', 'run_file', 'disable_pyimport', 'eval_js6',
    'translate_js6', 'PyJsException', 'get_file_contents',
    'write_file_contents', 'require'
]

from .base import PyJsException
from .evaljs import *
from .translators import parse as parse_js
from .node_import import require



from __future__ import unicode_literals
# Type Conversions. to_type. All must return PyJs subclass instance
from .simplex import *


def to_primitive(self, hint=None):
    if is_primitive(self):
        return self
    if hint is None and (self.Class == 'Number' or self.Class == 'Boolean'):
        # favour number for Class== Number or Boolean default = String
        hint = 'Number'
    return self.default_value(hint)


def to_boolean(self):
    typ = Type(self)
    if typ == 'Boolean':  # no need to convert
        return self
    elif typ == 'Null' or typ == 'Undefined':  # they are both always false
        return False
    elif typ == 'Number':  # false only for 0, and NaN
        return self and self == self  # test for nan (nan -> flase)
    elif typ == 'String':
        return bool(self)
    else:  # object -  always True
        return True


def to_number(self):
    typ = Type(self)
    if typ == 'Number':  # or self.Class=='Number':   # no need to convert
        return self
    elif typ == 'Null':  # null is 0
        return 0.
    elif typ == 'Undefined':  # undefined is NaN
        return NaN
    elif typ == 'Boolean':  # 1 for True 0 for false
        return float(self)
    elif typ == 'String':
        s = self.strip()  # Strip white space
        if not s:  # '' is simply 0
            return 0.
        if 'x' in s or 'X' in s[:3]:  # hex (positive only)
            try:  # try to convert
                num = int(s, 16)
            except ValueError:  # could not convert -> NaN
                return NaN
            return float(num)
        sign = 1  # get sign
        if s[0] in '+-':
            if s[0] == '-':
                sign = -1
            s = s[1:]
        if s == 'Infinity':  # Check for infinity keyword. 'NaN' will be NaN anyway.
            return sign * Infinity
        try:  # decimal try
            num = sign * float(s)  # Converted
        except ValueError:
            return NaN  # could not convert to decimal  > return NaN
        return float(num)
    else:  # object -  most likely it will be NaN.
        return to_number(to_primitive(self, 'Number'))


def to_string(self):
    typ = Type(self)
    if typ == 'String':
        return self
    elif typ == 'Null':
        return 'null'
    elif typ == 'Undefined':
        return 'undefined'
    elif typ == 'Boolean':
        return 'true' if self else 'false'
    elif typ == 'Number':  # or self.Class=='Number':
        return js_dtoa(self)
    else:  # object
        return to_string(to_primitive(self, 'String'))


def to_object(self, space):
    typ = Type(self)
    if typ == 'Object':
        return self
    elif typ == 'Boolean':  # Unsure ... todo check here
        return space.Boolean.create((self, ), space)
    elif typ == 'Number':  # ?
        return space.Number.create((self, ), space)
    elif typ == 'String':  # ?
        return space.String.create((self, ), space)
    elif typ == 'Null' or typ == 'Undefined':
        raise MakeError('TypeError',
                        'undefined or null can\'t be converted to object')
    else:
        raise RuntimeError()


def to_int32(self):
    num = to_number(self)
    if is_nan(num) or is_infinity(num):
        return 0
    int32 = int(num) % 2**32
    return int(int32 - 2**32 if int32 >= 2**31 else int32)


def to_int(self):
    num = to_number(self)
    if is_nan(num):
        return 0
    elif is_infinity(num):
        return 10**20 if num > 0 else -10**20
    return int(num)


def to_uint32(self):
    num = to_number(self)
    if is_nan(num) or is_infinity(num):
        return 0
    return int(num) % 2**32


def to_uint16(self):
    num = to_number(self)
    if is_nan(num) or is_infinity(num):
        return 0
    return int(num) % 2**16


def to_int16(self):
    num = to_number(self)
    if is_nan(num) or is_infinity(num):
        return 0
    int16 = int(num) % 2**16
    return int(int16 - 2**16 if int16 >= 2**15 else int16)


def cok(self):
    """Check object coercible"""
    if type(self) in (UNDEFINED_TYPE, NULL_TYPE):
        raise MakeError('TypeError',
                        'undefined or null can\'t be converted to object')



from __future__ import unicode_literals
from ..conversions import *
from ..func_utils import *
from ..base import is_data_descriptor
import six


def Object(this, args):
    val = get_arg(args, 0)
    if is_null(val) or is_undefined(val):
        return args.space.NewObject()
    return to_object(val, args.space)


def ObjectCreate(args, space):
    if len(args):
        val = get_arg(args, 0)
        if is_object(val):
            #  Implementation dependent, but my will simply return :)
            return val
        elif type(val) in (NUMBER_TYPE, STRING_TYPE, BOOLEAN_TYPE):
            return to_object(val, space)
    return space.NewObject()


class ObjectMethods:
    def getPrototypeOf(this, args):
        obj = get_arg(args, 0)
        if not is_object(obj):
            raise MakeError('TypeError',
                            'Object.getPrototypeOf called on non-object')
        return null if obj.prototype is None else obj.prototype

    def getOwnPropertyDescriptor(this, args):
        obj = get_arg(args, 0)
        prop = get_arg(args, 1)
        if not is_object(obj):
            raise MakeError(
                'TypeError',
                'Object.getOwnPropertyDescriptor called on non-object')
        desc = obj.own.get(to_string(prop))
        return convert_to_js_type(desc, args.space)

    def getOwnPropertyNames(this, args):
        obj = get_arg(args, 0)
        if not is_object(obj):
            raise MakeError(
                'TypeError',
                'Object.getOwnPropertyDescriptor called on non-object')
        return args.space.ConstructArray(obj.own.keys())

    def create(this, args):
        obj = get_arg(args, 0)
        if not (is_object(obj) or is_null(obj)):
            raise MakeError('TypeError',
                            'Object prototype may only be an Object or null')
        temp = args.space.NewObject()
        temp.prototype = None if is_null(obj) else obj
        if len(args) > 1 and not is_undefined(args[1]):
            if six.PY2:
                args.tup = (args[1], )
                ObjectMethods.defineProperties.__func__(temp, args)
            else:
                args.tup = (args[1], )
                ObjectMethods.defineProperties(temp, args)
        return temp

    def defineProperty(this, args):
        obj = get_arg(args, 0)
        prop = get_arg(args, 1)
        attrs = get_arg(args, 2)
        if not is_object(obj):
            raise MakeError('TypeError',
                            'Object.defineProperty called on non-object')
        name = to_string(prop)
        if not obj.define_own_property(name, ToPropertyDescriptor(attrs),
                                       False):
            raise MakeError('TypeError', 'Cannot redefine property: %s' % name)
        return obj

    def defineProperties(this, args):
        obj = get_arg(args, 0)
        properties = get_arg(args, 1)
        if not is_object(obj):
            raise MakeError('TypeError',
                            'Object.defineProperties called on non-object')
        props = to_object(properties, args.space)
        for k, v in props.own.items():
            if not v.get('enumerable'):
                continue
            desc = ToPropertyDescriptor(props.get(unicode(k)))
            if not obj.define_own_property(unicode(k), desc, False):
                raise MakeError('TypeError',
                                'Failed to define own property: %s' % k)
        return obj

    def seal(this, args):
        obj = get_arg(args, 0)
        if not is_object(obj):
            raise MakeError('TypeError', 'Object.seal called on non-object')
        for desc in obj.own.values():
            desc['configurable'] = False
        obj.extensible = False
        return obj

    def freeze(this, args):
        obj = get_arg(args, 0)
        if not is_object(obj):
            raise MakeError('TypeError', 'Object.freeze called on non-object')
        for desc in obj.own.values():
            desc['configurable'] = False
            if is_data_descriptor(desc):
                desc['writable'] = False
        obj.extensible = False
        return obj

    def preventExtensions(this, args):
        obj = get_arg(args, 0)
        if not is_object(obj):
            raise MakeError('TypeError',
                            'Object.preventExtensions on non-object')
        obj.extensible = False
        return obj

    def isSealed(this, args):
        obj = get_arg(args, 0)
        if not is_object(obj):
            raise MakeError('TypeError',
                            'Object.isSealed called on non-object')
        if obj.extensible:
            return False
        for desc in obj.own.values():
            if desc.get('configurable'):
                return False
        return True

    def isFrozen(this, args):
        obj = get_arg(args, 0)
        if not is_object(obj):
            raise MakeError('TypeError',
                            'Object.isFrozen called on non-object')
        if obj.extensible:
            return False
        for desc in obj.own.values():
            if desc.get('configurable'):
                return False
            if is_data_descriptor(desc) and desc.get('writable'):
                return False
        return True

    def isExtensible(this, args):
        obj = get_arg(args, 0)
        if not is_object(obj):
            raise MakeError('TypeError',
                            'Object.isExtensible called on non-object')
        return obj.extensible

    def keys(this, args):
        obj = get_arg(args, 0)
        if not is_object(obj):
            raise MakeError('TypeError', 'Object.keys called on non-object')
        return args.space.ConstructArray([
            unicode(e) for e, d in six.iteritems(obj.own)
            if d.get('enumerable')
        ])


# some utility functions:


def ToPropertyDescriptor(obj):  # page 38 (50 absolute)
    if not is_object(obj):
        raise MakeError('TypeError',
                        'Can\'t convert non-object to property descriptor')
    desc = {}
    if obj.has_property('enumerable'):
        desc['enumerable'] = to_boolean(obj.get('enumerable'))
    if obj.has_property('configurable'):
        desc['configurable'] = to_boolean(obj.get('configurable'))
    if obj.has_property('value'):
        desc['value'] = obj.get('value')
    if obj.has_property('writable'):
        desc['writable'] = to_boolean(obj.get('writable'))
    if obj.has_property('get'):
        cand = obj.get('get')
        if not (is_undefined(cand) or is_callable(cand)):
            raise MakeError(
                'TypeError',
                'Invalid getter (it has to be a function or undefined)')
        desc['get'] = cand
    if obj.has_property('set'):
        cand = obj.get('set')
        if not (is_undefined(cand) or is_callable(cand)):
            raise MakeError(
                'TypeError',
                'Invalid setter (it has to be a function or undefined)')
        desc['set'] = cand
    if ('get' in desc or 'set' in desc) and ('value' in desc
                                             or 'writable' in desc):
        raise MakeError(
            'TypeError',
            'Invalid property.  A property cannot both have accessors and be writable or have a value.'
        )
    return desc



from .simplex import *
from .conversions import *

import six
if six.PY3:
    basestring = str
    long = int
    xrange = range
    unicode = str


def get_arg(arguments, n):
    if len(arguments) <= n:
        return undefined
    return arguments[n]


def ensure_js_types(args, space=None):
    return tuple(convert_to_js_type(e, space=space) for e in args)


def convert_to_js_type(e, space=None):
    t = type(e)
    if is_js_type(e):
        return e
    if t in (int, long, float):
        return float(e)
    elif isinstance(t, basestring):
        return unicode(t)
    elif t in (list, tuple):
        if space is None:
            raise MakeError(
                'TypeError',
                'Actually an internal error, could not convert to js type because space not specified'
            )
        return space.ConstructArray(ensure_js_types(e, space=space))
    elif t == dict:
        if space is None:
            raise MakeError(
                'TypeError',
                'Actually an internal error, could not convert to js type because space not specified'
            )
        new = {}
        for k, v in e.items():
            new[to_string(convert_to_js_type(k, space))] = convert_to_js_type(
                v, space)
        return space.ConstructObject(new)
    else:
        raise MakeError('TypeError', 'Could not convert to js type!')


def is_js_type(e):
    if type(e) in PRIMITIVES:
        return True
    elif hasattr(e, 'Class') and hasattr(e, 'value'):  # not perfect but works
        return True
    else:
        return False


# todo optimise these 2!
def js_array_to_tuple(arr):
    length = to_uint32(arr.get(u'length'))
    return tuple(arr.get(unicode(e)) for e in xrange(length))


def js_array_to_list(arr):
    length = to_uint32(arr.get(u'length'))
    return [arr.get(unicode(e)) for e in xrange(length)]


def js_arr_length(arr):
    return to_uint32(arr.get(u'length'))



from ..base import *
from ..conversions import *
from ..func_utils import *
from pyjsparser import parse
from ..byte_trans import ByteCodeGenerator, Code


def Function(this, args):
    # convert arguments to python list of strings
    a = list(map(to_string, tuple(args)))
    _body = u';'
    _args = ()
    if len(a):
        _body = u'%s;' % a[-1]
        _args = a[:-1]
    return executable_function(_body, _args, args.space, global_context=True)


def executable_function(_body, _args, space, global_context=True):
    func_str = u'(function (%s) { ; %s ; });' % (u', '.join(_args), _body)

    co = executable_code(
        code_str=func_str, space=space, global_context=global_context)
    return co()


# you can use this one lovely piece of function to compile and execute code on the fly! Watch out though as it may generate lots of code.
# todo tape cleanup? we dont know which pieces are needed and which are not so rather impossible without smarter machinery something like GC,
# a one solution would be to have a separate tape for functions
def executable_code(code_str, space, global_context=True):
    # parse first to check if any SyntaxErrors
    parsed = parse(code_str)

    old_tape_len = len(space.byte_generator.exe.tape)
    space.byte_generator.record_state()
    start = space.byte_generator.exe.get_new_label()
    skip = space.byte_generator.exe.get_new_label()
    space.byte_generator.emit('JUMP', skip)
    space.byte_generator.emit('LABEL', start)
    space.byte_generator.emit(parsed)
    space.byte_generator.emit('NOP')
    space.byte_generator.emit('LABEL', skip)
    space.byte_generator.emit('NOP')
    space.byte_generator.restore_state()

    space.byte_generator.exe.compile(
        start_loc=old_tape_len
    )  # dont read the code from the beginning, dont be stupid!

    ctx = space.GlobalObj if global_context else space.exe.current_ctx

    def ex_code():
        ret, status, token = space.byte_generator.exe.execute_fragment_under_context(
            ctx, start, skip)
        # todo Clean up the tape!
        # this is NOT a way to do that because the fragment may contain the executable code! We dont want to remove it
        #del space.byte_generator.exe.tape[old_tape_len:]
        if status == 0:
            return ret
        elif status == 3:
            raise token
        else:
            raise RuntimeError(
                'Unexpected return status during JIT execution: %d' % status)

    return ex_code


def _eval(this, args):
    code_str = to_string(get_arg(args, 0))
    return executable_code(code_str, args.space, global_context=True)()


def log(this, args):
    print(' '.join(map(to_string, args)))
    return undefined



from ..base import *

CONSTS = {
    'prototype': NumberPrototype,
    'MAX_VALUE': 1.7976931348623157e308,
    'MIN_VALUE': 5.0e-324,
    'NaN': NaN,
    'NEGATIVE_INFINITY': float('-inf'),
    'POSITIVE_INFINITY': float('inf')
}

fill_in_props(Number, CONSTS, {
    'enumerable': False,
    'writable': False,
    'configurable': False
})

NumberPrototype.define_own_property('constructor', {
    'value': Number,
    'enumerable': False,
    'writable': True,
    'configurable': True
})



from ..base import *
from .time_helpers import *

TZ_OFFSET = (time.altzone // 3600)
ABS_OFFSET = abs(TZ_OFFSET)
TZ_NAME = time.tzname[1]
ISO_FORMAT = '%s-%s-%sT%s:%s:%s.%sZ'


@Js
def Date(year, month, date, hours, minutes, seconds, ms):
    return now().to_string()


Date.Class = 'Date'


def now():
    return PyJsDate(int(time.time() * 1000), prototype=DatePrototype)


@Js
def UTC(year, month, date, hours, minutes, seconds, ms):  # todo complete this
    args = arguments
    y = args[0].to_number()
    m = args[1].to_number()
    l = len(args)
    dt = args[2].to_number() if l > 2 else Js(1)
    h = args[3].to_number() if l > 3 else Js(0)
    mi = args[4].to_number() if l > 4 else Js(0)
    sec = args[5].to_number() if l > 5 else Js(0)
    mili = args[6].to_number() if l > 6 else Js(0)
    if not y.is_nan() and 0 <= y.value <= 99:
        y = y + Js(1900)
    return TimeClip(MakeDate(MakeDay(y, m, dt), MakeTime(h, mi, sec, mili)))


@Js
def parse(string):
    return PyJsDate(
        TimeClip(parse_date(string.to_string().value)),
        prototype=DatePrototype)


Date.define_own_property('now', {
    'value': Js(now),
    'enumerable': False,
    'writable': True,
    'configurable': True
})

Date.define_own_property('parse', {
    'value': parse,
    'enumerable': False,
    'writable': True,
    'configurable': True
})

Date.define_own_property('UTC', {
    'value': UTC,
    'enumerable': False,
    'writable': True,
    'configurable': True
})


class PyJsDate(PyJs):
    Class = 'Date'
    extensible = True

    def __init__(self, value, prototype=None):
        self.value = value
        self.own = {}
        self.prototype = prototype

    # todo fix this problematic datetime part
    def to_local_dt(self):
        return datetime.datetime(1970, 1, 1) + datetime.timedelta(
            seconds=UTCToLocal(self.value) // 1000)

    def to_utc_dt(self):
        return datetime.datetime(1970, 1, 1) + datetime.timedelta(
            seconds=self.value // 1000)

    def local_strftime(self, pattern):
        if self.value is NaN:
            return 'Invalid Date'
        try:
            dt = self.to_local_dt()
        except:
            raise MakeError(
                'TypeError',
                'unsupported date range. Will fix in future versions')
        try:
            return dt.strftime(pattern)
        except:
            raise MakeError(
                'TypeError',
                'Could not generate date string from this date (limitations of python.datetime)'
            )

    def utc_strftime(self, pattern):
        if self.value is NaN:
            return 'Invalid Date'
        try:
            dt = self.to_utc_dt()
        except:
            raise MakeError(
                'TypeError',
                'unsupported date range. Will fix in future versions')
        try:
            return dt.strftime(pattern)
        except:
            raise MakeError(
                'TypeError',
                'Could not generate date string from this date (limitations of python.datetime)'
            )


def parse_date(py_string):  # todo support all date string formats
    date_formats = (
        "%Y-%m-%d",
        "%m/%d/%Y",
        "%b %d %Y",
    )
    # Supports these hour formats and with or hour.
    hour_formats = (
        "T%H:%M:%S.%f",
        "T%H:%M:%S",
    ) + ('',)
    # Supports with or without Z indicator.
    z_formats = ("Z",) + ('',)
    supported_formats = [
        d + t + z
        for d in date_formats
        for t in hour_formats
        for z in z_formats
    ]
    for date_format in supported_formats:
        try:
            dt = datetime.datetime.strptime(py_string, date_format)
        except ValueError:
            continue
        else:
            return MakeDate(
                MakeDay(Js(dt.year), Js(dt.month - 1), Js(dt.day)),
                MakeTime(
                    Js(dt.hour), Js(dt.minute), Js(dt.second),
                    Js(dt.microsecond // 1000)))

    raise MakeError(
        'TypeError',
        'Could not parse date %s - unsupported date format. Currently only supported formats are RFC3339 utc, ISO Date, Short Date, and Long Date. Sorry!'
        % py_string)


def date_constructor(*args):
    if len(args) >= 2:
        return date_constructor2(*args)
    elif len(args) == 1:
        return date_constructor1(args[0])
    else:
        return date_constructor0()


def date_constructor0():
    return now()


def date_constructor1(value):
    v = value.to_primitive()
    if v._type() == 'String':
        v = parse_date(v.value)
    else:
        v = v.to_int()
    return PyJsDate(TimeClip(v), prototype=DatePrototype)


def date_constructor2(*args):
    y = args[0].to_number()
    m = args[1].to_number()
    l = len(args)
    dt = args[2].to_number() if l > 2 else Js(1)
    h = args[3].to_number() if l > 3 else Js(0)
    mi = args[4].to_number() if l > 4 else Js(0)
    sec = args[5].to_number() if l > 5 else Js(0)
    mili = args[6].to_number() if l > 6 else Js(0)
    if not y.is_nan() and 0 <= y.value <= 99:
        y = y + Js(1900)
    t = TimeClip(
        LocalToUTC(MakeDate(MakeDay(y, m, dt), MakeTime(h, mi, sec, mili))))
    return PyJsDate(t, prototype=DatePrototype)


Date.create = date_constructor

DatePrototype = PyJsDate(float('nan'), prototype=ObjectPrototype)


def check_date(obj):
    if obj.Class != 'Date':
        raise MakeError('TypeError', 'this is not a Date object')


class DateProto:
    def toString():
        check_date(this)
        if this.value is NaN:
            return 'Invalid Date'
        offset = (UTCToLocal(this.value) - this.value) // msPerHour
        return this.local_strftime(
            '%a %b %d %Y %H:%M:%S GMT') + '%s00 (%s)' % (pad(
                offset, 2, True), GetTimeZoneName(this.value))

    def toDateString():
        check_date(this)
        return this.local_strftime('%d %B %Y')

    def toTimeString():
        check_date(this)
        return this.local_strftime('%H:%M:%S')

    def toLocaleString():
        check_date(this)
        return this.local_strftime('%d %B %Y %H:%M:%S')

    def toLocaleDateString():
        check_date(this)
        return this.local_strftime('%d %B %Y')

    def toLocaleTimeString():
        check_date(this)
        return this.local_strftime('%H:%M:%S')

    def valueOf():
        check_date(this)
        return this.value

    def getTime():
        check_date(this)
        return this.value

    def getFullYear():
        check_date(this)
        if this.value is NaN:
            return NaN
        return YearFromTime(UTCToLocal(this.value))

    def getUTCFullYear():
        check_date(this)
        if this.value is NaN:
            return NaN
        return YearFromTime(this.value)

    def getMonth():
        check_date(this)
        if this.value is NaN:
            return NaN
        return MonthFromTime(UTCToLocal(this.value))

    def getDate():
        check_date(this)
        if this.value is NaN:
            return NaN
        return DateFromTime(UTCToLocal(this.value))

    def getUTCMonth():
        check_date(this)
        if this.value is NaN:
            return NaN
        return MonthFromTime(this.value)

    def getUTCDate():
        check_date(this)
        if this.value is NaN:
            return NaN
        return DateFromTime(this.value)

    def getDay():
        check_date(this)
        if this.value is NaN:
            return NaN
        return WeekDay(UTCToLocal(this.value))

    def getUTCDay():
        check_date(this)
        if this.value is NaN:
            return NaN
        return WeekDay(this.value)

    def getHours():
        check_date(this)
        if this.value is NaN:
            return NaN
        return HourFromTime(UTCToLocal(this.value))

    def getUTCHours():
        check_date(this)
        if this.value is NaN:
            return NaN
        return HourFromTime(this.value)

    def getMinutes():
        check_date(this)
        if this.value is NaN:
            return NaN
        return MinFromTime(UTCToLocal(this.value))

    def getUTCMinutes():
        check_date(this)
        if this.value is NaN:
            return NaN
        return MinFromTime(this.value)

    def getSeconds():
        check_date(this)
        if this.value is NaN:
            return NaN
        return SecFromTime(UTCToLocal(this.value))

    def getUTCSeconds():
        check_date(this)
        if this.value is NaN:
            return NaN
        return SecFromTime(this.value)

    def getMilliseconds():
        check_date(this)
        if this.value is NaN:
            return NaN
        return msFromTime(UTCToLocal(this.value))

    def getUTCMilliseconds():
        check_date(this)
        if this.value is NaN:
            return NaN
        return msFromTime(this.value)

    def getTimezoneOffset():
        check_date(this)
        if this.value is NaN:
            return NaN
        return (this.value - UTCToLocal(this.value)) // 60000

    def setTime(time):
        check_date(this)
        this.value = TimeClip(time.to_number().to_int())
        return this.value

    def setMilliseconds(ms):
        check_date(this)
        t = UTCToLocal(this.value)
        tim = MakeTime(
            Js(HourFromTime(t)), Js(MinFromTime(t)), Js(SecFromTime(t)), ms)
        u = TimeClip(LocalToUTC(MakeDate(Day(t), tim)))
        this.value = u
        return u

    def setUTCMilliseconds(ms):
        check_date(this)
        t = this.value
        tim = MakeTime(
            Js(HourFromTime(t)), Js(MinFromTime(t)), Js(SecFromTime(t)), ms)
        u = TimeClip(MakeDate(Day(t), tim))
        this.value = u
        return u

    def setSeconds(sec, ms=None):
        check_date(this)
        t = UTCToLocal(this.value)
        s = sec.to_number()
        if not ms is None: milli = Js(msFromTime(t))
        else: milli = ms.to_number()
        date = MakeDate(
            Day(t), MakeTime(Js(HourFromTime(t)), Js(MinFromTime(t)), s, milli))
        u = TimeClip(LocalToUTC(date))
        this.value = u
        return u

    def setUTCSeconds(sec, ms=None):
        check_date(this)
        t = this.value
        s = sec.to_number()
        if not ms is None: milli = Js(msFromTime(t))
        else: milli = ms.to_number()
        date = MakeDate(
            Day(t), MakeTime(Js(HourFromTime(t)), Js(MinFromTime(t)), s, milli))
        v = TimeClip(date)
        this.value = v
        return v

    def setMinutes(min, sec=None, ms=None):
        check_date(this)
        t = UTCToLocal(this.value)
        m = min.to_number()
        if not sec is None: s = Js(SecFromTime(t))
        else: s = sec.to_number()
        if not ms is None: milli = Js(msFromTime(t))
        else: milli = ms.to_number()
        date = MakeDate(Day(t), MakeTime(Js(HourFromTime(t)), m, s, milli))
        u = TimeClip(LocalToUTC(date))
        this.value = u
        return u

    def setUTCMinutes(min, sec=None, ms=None):
        check_date(this)
        t = this.value
        m = min.to_number()
        if not sec is None: s = Js(SecFromTime(t))
        else: s = sec.to_number()
        if not ms is None: milli = Js(msFromTime(t))
        else: milli = ms.to_number()
        date = MakeDate(Day(t), MakeTime(Js(HourFromTime(t)), m, s, milli))
        v = TimeClip(date)
        this.value = v
        return v

    def setHours(hour, min=None, sec=None, ms=None):
        check_date(this)
        t = UTCToLocal(this.value)
        h = hour.to_number()
        if not min is None: m = Js(MinFromTime(t))
        else: m = min.to_number()
        if not sec is None: s = Js(SecFromTime(t))
        else: s = sec.to_number()
        if not ms is None: milli = Js(msFromTime(t))
        else: milli = ms.to_number()
        date = MakeDate(Day(t), MakeTime(h, m, s, milli))
        u = TimeClip(LocalToUTC(date))
        this.value = u
        return u

    def setUTCHours(hour, min=None, sec=None, ms=None):
        check_date(this)
        t = this.value
        h = hour.to_number()
        if not min is None: m = Js(MinFromTime(t))
        else: m = min.to_number()
        if not sec is None: s = Js(SecFromTime(t))
        else: s = sec.to_number()
        if not ms is None: milli = Js(msFromTime(t))
        else: milli = ms.to_number()
        date = MakeDate(Day(t), MakeTime(h, m, s, milli))
        v = TimeClip(date)
        this.value = v
        return v

    def setDate(date):
        check_date(this)
        t = UTCToLocal(this.value)
        dt = date.to_number()
        newDate = MakeDate(
            MakeDay(Js(YearFromTime(t)), Js(MonthFromTime(t)), dt), TimeWithinDay(t))
        u = TimeClip(LocalToUTC(newDate))
        this.value = u
        return u

    def setUTCDate(date):
        check_date(this)
        t = this.value
        dt = date.to_number()
        newDate = MakeDate(
            MakeDay(Js(YearFromTime(t)), Js(MonthFromTime(t)), dt), TimeWithinDay(t))
        v = TimeClip(newDate)
        this.value = v
        return v

    def setMonth(month, date=None):
        check_date(this)
        t = UTCToLocal(this.value)
        m = month.to_number()
        if not date is None: dt = Js(DateFromTime(t))
        else: dt = date.to_number()
        newDate = MakeDate(
            MakeDay(Js(YearFromTime(t)), m, dt), TimeWithinDay(t))
        u = TimeClip(LocalToUTC(newDate))
        this.value = u
        return u

    def setUTCMonth(month, date=None):
        check_date(this)
        t = this.value
        m = month.to_number()
        if not date is None: dt = Js(DateFromTime(t))
        else: dt = date.to_number()
        newDate = MakeDate(
            MakeDay(Js(YearFromTime(t)), m, dt), TimeWithinDay(t))
        v = TimeClip(newDate)
        this.value = v
        return v

    def setFullYear(year, month=None, date=None):
        check_date(this)
        if not this.value is NaN: t = UTCToLocal(this.value)
        else: t = 0
        y = year.to_number()
        if not month is None: m = Js(MonthFromTime(t))
        else: m = month.to_number()
        if not date is None: dt = Js(DateFromTime(t))
        else: dt = date.to_number()
        newDate = MakeDate(
            MakeDay(y, m, dt), TimeWithinDay(t))
        u = TimeClip(LocalToUTC(newDate))
        this.value = u
        return u

    def setUTCFullYear(year, month=None, date=None):
        check_date(this)
        if not this.value is NaN: t = UTCToLocal(this.value)
        else: t = 0
        y = year.to_number()
        if not month is None: m = Js(MonthFromTime(t))
        else: m = month.to_number()
        if not date is None: dt = Js(DateFromTime(t))
        else: dt = date.to_number()
        newDate = MakeDate(
            MakeDay(y, m, dt), TimeWithinDay(t))
        v = TimeClip(newDate)
        this.value = v
        return v

    def toUTCString():
        check_date(this)
        return this.utc_strftime('%d %B %Y %H:%M:%S')

    def toISOString():
        check_date(this)
        t = this.value
        year = YearFromTime(t)
        month, day, hour, minute, second, milli = pad(
            MonthFromTime(t) + 1), pad(DateFromTime(t)), pad(
                HourFromTime(t)), pad(MinFromTime(t)), pad(
                    SecFromTime(t)), pad(msFromTime(t))
        return ISO_FORMAT % (unicode(year) if 0 <= year <= 9999 else pad(
            year, 6, True), month, day, hour, minute, second, milli)

    def toJSON(key):
        o = this.to_object()
        tv = o.to_primitive('Number')
        if tv.Class == 'Number' and not tv.is_finite():
            return this.null
        toISO = o.get('toISOString')
        if not toISO.is_callable():
            raise this.MakeError('TypeError', 'toISOString is not callable')
        return toISO.call(o, ())


def pad(num, n=2, sign=False):
    '''returns n digit string representation of the num'''
    s = unicode(abs(num))
    if len(s) < n:
        s = '0' * (n - len(s)) + s
    if not sign:
        return s
    if num >= 0:
        return '+' + s
    else:
        return '-' + s


fill_prototype(DatePrototype, DateProto, default_attrs)

Date.define_own_property(
    'prototype', {
        'value': DatePrototype,
        'enumerable': False,
        'writable': False,
        'configurable': False
    })

DatePrototype.define_own_property('constructor', {
    'value': Date,
    'enumerable': False,
    'writable': True,
    'configurable': True
})



class RegExpPrototype:
    def toString():
        flags = u''
        try:
            if this.glob:
                flags += u'g'
            if this.ignore_case:
                flags += u'i'
            if this.multiline:
                flags += u'm'
        except:
            pass
        v = this.value if this.value else '(?:)'
        return u'/%s/' % v + flags

    def test(string):
        return Exec(this, string) is not this.null

    def exec2(string
              ):  # will be changed to exec in base.py. cant name it exec here
        return Exec(this, string)


def Exec(this, string):
    if this.Class != 'RegExp':
        raise this.MakeError('TypeError',
                             'RegExp.prototype.exec is not generic!')
    string = string.to_string()
    length = len(string)
    i = this.get('lastIndex').to_int() if this.glob else 0
    matched = False
    while not matched:
        if i < 0 or i > length:
            this.put('lastIndex', this.Js(0))
            return this.null
        matched = this.match(string.value, i)
        i += 1
    start, end = matched.span()  #[0]+i-1, matched.span()[1]+i-1
    if this.glob:
        this.put('lastIndex', this.Js(end))
    arr = this.Js(
        [this.Js(e) for e in [matched.group()] + list(matched.groups())])
    arr.put('index', this.Js(start))
    arr.put('input', string)
    return arr



from ..conversions import *
from ..func_utils import *


def Array(this, args):
    return ArrayConstructor(args, args.space)


def ArrayConstructor(args, space):
    if len(args) == 1:
        l = get_arg(args, 0)
        if type(l) == float:
            if to_uint32(l) == l:
                return space.NewArray(l)
            else:
                raise MakeError(
                    'RangeError',
                    'Invalid length specified for Array constructor (must be uint32)'
                )
        else:
            return space.ConstructArray([l])
    else:
        return space.ConstructArray(list(args))


def isArray(this, args):
    x = get_arg(args, 0)
    return is_object(x) and x.Class == u'Array'



from __future__ import unicode_literals

from ..conversions import *
from ..func_utils import *


class RegExpPrototype:
    def toString(this, args):
        flags = u''
        try:
            if this.glob:
                flags += u'g'
            if this.ignore_case:
                flags += u'i'
            if this.multiline:
                flags += u'm'
        except:
            pass
        try:
            v = this.value if this.value else u'(?:)'
        except:
            v = u'(?:)'
        return u'/%s/' % v + flags

    def test(this, args):
        string = get_arg(args, 0)
        return RegExpExec(this, string, args.space) is not null

    def _exec(
            this, args
    ):  # will be changed to exec in base.py. cant name it exec here...
        string = get_arg(args, 0)
        return RegExpExec(this, string, args.space)


def RegExpExec(this, string, space):
    if GetClass(this) != 'RegExp':
        raise MakeError('TypeError', 'RegExp.prototype.exec is not generic!')
    string = to_string(string)
    length = len(string)
    i = to_int(this.get('lastIndex')) if this.glob else 0
    matched = False
    while not matched:
        if i < 0 or i > length:
            this.put('lastIndex', 0.)
            return null
        matched = this.match(string, i)
        i += 1
    start, end = matched.span()  #[0]+i-1, matched.span()[1]+i-1
    if this.glob:
        this.put('lastIndex', float(end))
    arr = convert_to_js_type(
        [matched.group()] + list(matched.groups()), space=space)
    arr.put('index', float(start))
    arr.put('input', unicode(string))
    return arr



__all__ = ['fix_js_args']

import types
from collections import namedtuple
import opcode
import six
import sys
import dis

if six.PY3:
    xrange = range
    chr = lambda x: x

# Opcode constants used for comparison and replacecment
LOAD_FAST = opcode.opmap['LOAD_FAST']
LOAD_GLOBAL = opcode.opmap['LOAD_GLOBAL']
STORE_FAST = opcode.opmap['STORE_FAST']


def fix_js_args(func):
    '''Use this function when unsure whether func takes this and arguments as its last 2 args.
       It will append 2 args if it does not.'''
    fcode = six.get_function_code(func)
    fargs = fcode.co_varnames[fcode.co_argcount - 2:fcode.co_argcount]
    if fargs == ('this', 'arguments') or fargs == ('arguments', 'var'):
        return func
    code = append_arguments(six.get_function_code(func), ('this', 'arguments'))

    result = types.FunctionType(
        code,
        six.get_function_globals(func),
        func.__name__,
        closure=six.get_function_closure(func))
    return result


def append_arguments(code_obj, new_locals):
    co_varnames = code_obj.co_varnames  # Old locals
    co_names = code_obj.co_names  # Old globals
    new_args = tuple(e for e in new_locals if e not in co_names)
    co_names += new_args
    co_argcount = code_obj.co_argcount  # Argument count
    co_code = code_obj.co_code  # The actual bytecode as a string

    # Make one pass over the bytecode to identify names that should be
    # left in code_obj.co_names.
    not_removed = set(opcode.hasname) - set([LOAD_GLOBAL])
    saved_names = set()
    for inst in instructions(code_obj):
        if inst[0] in not_removed:
            saved_names.add(co_names[inst[1]])

    # Build co_names for the new code object. This should consist of
    # globals that were only accessed via LOAD_GLOBAL
    names = tuple(
        name for name in co_names if name not in set(new_locals) - saved_names)

    # Build a dictionary that maps the indices of the entries in co_names
    # to their entry in the new co_names
    name_translations = dict(
        (co_names.index(name), i) for i, name in enumerate(names))

    # Build co_varnames for the new code object. This should consist of
    # the entirety of co_varnames with new_locals spliced in after the
    # arguments
    new_locals_len = len(new_locals)
    varnames = (
        co_varnames[:co_argcount] + new_locals + co_varnames[co_argcount:])

    # Build the dictionary that maps indices of entries in the old co_varnames
    # to their indices in the new co_varnames
    range1, range2 = xrange(co_argcount), xrange(co_argcount, len(co_varnames))
    varname_translations = dict((i, i) for i in range1)
    varname_translations.update((i, i + new_locals_len) for i in range2)

    # Build the dictionary that maps indices of deleted entries of co_names
    # to their indices in the new co_varnames
    names_to_varnames = dict(
        (co_names.index(name), varnames.index(name)) for name in new_locals)

    is_new_bytecode = sys.version_info >= (3, 11)
    # Now we modify the actual bytecode
    modified = []
    drop_future_cache = False
    for inst in instructions(code_obj):
        if is_new_bytecode and inst.opname == "CACHE":
            assert inst.arg == 0
            if not drop_future_cache:
                modified.extend(write_instruction(inst.opcode, inst.arg))
            else:
                # We need to inject NOOP to not break jumps :(
                modified.extend(write_instruction(dis.opmap["NOP"], 0))

            continue
        op, arg = inst.opcode, inst.arg
        # If the instruction is a LOAD_GLOBAL, we have to check to see if
        # it's one of the globals that we are replacing. Either way,
        # update its arg using the appropriate dict.
        drop_future_cache = False
        if inst.opcode == LOAD_GLOBAL:
            idx = inst.arg
            if is_new_bytecode:
                idx = idx // 2
            if idx in names_to_varnames:
                op = LOAD_FAST
                arg = names_to_varnames[idx]
                # Cache is not present after LOAD_FAST and needs to be removed.
                drop_future_cache = True
            elif idx in name_translations:
                tgt = name_translations[idx]
                if is_new_bytecode:
                    tgt = 2*tgt + (inst.arg % 2)
                arg = tgt
            else:
                raise(ValueError("a name was lost in translation last instruction %s" % str(inst)))
        # If it accesses co_varnames or co_names then update its argument.
        elif inst.opcode in opcode.haslocal:
            arg = varname_translations[inst.arg]
        elif inst.opcode in opcode.hasname:
            # for example STORE_GLOBAL
            arg = name_translations[inst.arg]
        elif is_new_bytecode and inst.opcode in opcode.hasfree:
            # Python 3.11+ adds refs at the end (after locals), for whatever reason...
            if inst.argval not in code_obj.co_varnames[:code_obj.co_argcount]:  # we do not need to remap existing arguments, they are not shifted by new ones.
                arg = inst.arg + len(new_locals)
        modified.extend(write_instruction(op, arg))
    if six.PY2:
        code = ''.join(modified)
        args = (co_argcount + new_locals_len,
                code_obj.co_nlocals + new_locals_len, code_obj.co_stacksize,
                code_obj.co_flags, code, code_obj.co_consts, names, varnames,
                code_obj.co_filename, code_obj.co_name,
                code_obj.co_firstlineno, code_obj.co_lnotab,
                code_obj.co_freevars, code_obj.co_cellvars)
    else:
        code = bytes(modified)
        args = (co_argcount + new_locals_len, 0,
                code_obj.co_nlocals + new_locals_len, code_obj.co_stacksize,
                code_obj.co_flags, code, code_obj.co_consts, names, varnames,
                code_obj.co_filename, code_obj.co_name,
                code_obj.co_firstlineno, code_obj.co_lnotab,
                code_obj.co_freevars, code_obj.co_cellvars)
    # Done modifying codestring - make the code object
    if hasattr(code_obj, "replace"):
        # Python 3.8+
        code_obj = code_obj.replace(
            co_argcount=co_argcount + new_locals_len,
            co_nlocals=code_obj.co_nlocals + new_locals_len,
            co_code=code,
            co_names=names,
            co_varnames=varnames)
        return code_obj
    else:
        return types.CodeType(*args)


def instructions(code_obj, show_cache=True):
    if sys.version_info >= (3, 11):
        # Python 3.11 introduced "cache instructions", hidden by default.
        for inst in dis.Bytecode(code_obj, show_caches=show_cache):
            yield inst
    elif sys.version_info >= (3, 4): # easy for python 3.4+
        for inst in dis.Bytecode(code_obj):
            yield inst
    else:
        # otherwise we have to manually parse
        code = code_obj.co_code
        NewInstruction = namedtuple('Instruction', ('opcode', 'arg'))
        if six.PY2:
            code = map(ord, code)
        i, L = 0, len(code)
        extended_arg = 0
        while i < L:
            op = code[i]
            i += 1
            if op < opcode.HAVE_ARGUMENT:
                yield NewInstruction(op, None)
                continue
            oparg = code[i] + (code[i + 1] << 8) + extended_arg
            extended_arg = 0
            i += 2
            if op == opcode.EXTENDED_ARG:
                extended_arg = oparg << 16
                continue
            yield NewInstruction(op, oparg)


def write_instruction(op, arg):
    if sys.version_info < (3, 6):
        if arg is None:
            return [chr(op)]
        elif arg <= 65536:
            return [chr(op), chr(arg & 255), chr((arg >> 8) & 255)]
        elif arg <= 4294967296:
            return [
                chr(opcode.EXTENDED_ARG),
                chr((arg >> 16) & 255),
                chr((arg >> 24) & 255),
                chr(op),
                chr(arg & 255),
                chr((arg >> 8) & 255)
            ]
        else:
            raise ValueError("Invalid oparg: {0} is too large".format(arg))
    else:  # python 3.6+ uses wordcode instead of bytecode and they already supply all the EXTENDEND_ARG ops :)
        if arg is None:
            return [chr(op), 0]
        return [chr(op), arg & 255]
        # the code below is for case when extended args are to be determined automatically
        # if op == opcode.EXTENDED_ARG:
        #     return []  # this will be added automatically
        # elif arg < 1 << 8:
        #     return [chr(op), arg]
        # elif arg < 1 << 32:
        #     subs = [1<<24, 1<<16, 1<<8]  # allowed op extension sizes
        #     for sub in subs:
        #         if arg >= sub:
        #             fit = int(arg / sub)
        #             return [chr(opcode.EXTENDED_ARG), fit]  + write_instruction(op, arg - fit * sub)
        # else:
        #     raise ValueError("Invalid oparg: {0} is too large".format(oparg))



def check(code_obj):
    old_bytecode = code_obj.co_code
    insts = list(instructions(code_obj))

    pos_to_inst = {}
    bytelist = []

    for inst in insts:
        pos_to_inst[len(bytelist)] = inst
        bytelist.extend(write_instruction(inst.opcode, inst.arg))
    if six.PY2:
        new_bytecode = ''.join(bytelist)
    else:
        new_bytecode = bytes(bytelist)
    if new_bytecode != old_bytecode:
        print(new_bytecode)
        print(old_bytecode)
        for i in range(min(len(new_bytecode), len(old_bytecode))):
            if old_bytecode[i] != new_bytecode[i]:
                while 1:
                    if i in pos_to_inst:
                        print(pos_to_inst[i])
                        print(pos_to_inst[i - 2])
                        print(list(map(chr, old_bytecode))[i - 4:i + 8])
                        print(bytelist[i - 4:i + 8])
                        break
            raise RuntimeError(
                'Your python version made changes to the bytecode')




def signature(func):
    code_obj = six.get_function_code(func)
    return (code_obj.co_nlocals, code_obj.co_argcount, code_obj.co_nlocals, code_obj.co_stacksize,
    code_obj.co_flags, code_obj.co_names, code_obj.co_varnames,
    code_obj.co_filename,
    code_obj.co_freevars, code_obj.co_cellvars)

check(six.get_function_code(check))



def compare_func(fake_func, gt_func):
    print(signature(fake_func))
    print(signature(gt_func))
    assert signature(fake_func) == signature(gt_func)
    fake_ins = list(instructions(six.get_function_code(fake_func), show_cache=False))
    real_ins = list(instructions(six.get_function_code(gt_func), show_cache=False))
    offset = 0
    pos = 0
    for e in fake_ins:
        if e.opname == "NOP":
            offset += 1  # ignore NOPs that are inserted in place of old cache.
        else:
            real = real_ins[pos]
            fake = e
            print("POS %d OFFSET: %d   FAKE VS REAL" % (pos, offset))
            print(fake)
            print(real)
            assert fake.opcode == real.opcode
            if fake.opcode in dis.hasjabs or fake.opcode in dis.hasjrel:
                pass
            else:
                assert fake.arg == real.arg
                assert fake.argval == real.argval or fake.opname in ["LOAD_CONST"]
                assert fake.is_jump_target == real.is_jump_target

            pos += 1
    assert pos == len(real_ins), (pos, len(real_ins))
    print("DONE, looks good.")


if __name__ == '__main__':
    import faulthandler

    faulthandler.enable()

    def func(cmpfn):
        if not this.Class in ('Array', 'Arguments'):
            return this.to_object()  # do nothing
        arr = []
        for i in xrange(len(this)):
            arr.append(this.get(six.text_type(i)))

        if not arr:
            return this
        if not cmpfn.is_callable():
            cmpfn = None
        cmp = lambda a, b: sort_compare(a, b, cmpfn)
        if six.PY3:
            key = functools.cmp_to_key(cmp)
            arr.sort(key=key)
        else:
            arr.sort(cmp=cmp)
        for i in xrange(len(arr)):
            this.put(six.text_type(i), arr[i])

        return this


    def func_gt(cmpfn, this, arguments):
        if not this.Class in ('Array', 'Arguments'):
            return this.to_object()  # do nothing
        arr = []
        for i in xrange(len(this)):
            arr.append(this.get(six.text_type(i)))

        if not arr:
            return this
        if not cmpfn.is_callable():
            cmpfn = None
        cmp = lambda a, b: sort_compare(a, b, cmpfn)
        if six.PY3:
            key = functools.cmp_to_key(cmp)
            arr.sort(key=key)
        else:
            arr.sort(cmp=cmp)
        for i in xrange(len(arr)):
            this.put(six.text_type(i), arr[i])

        return this


    func2 = fix_js_args(func)
    compare_func(func2, func_gt)



import pyjsparser
import pyjsparser.parser
from . import translating_nodes

import hashlib
import re

# Enable Js2Py exceptions and pyimport in parser
pyjsparser.parser.ENABLE_PYIMPORT = True

# the re below is how we'll recognise numeric constants.
# it finds any 'simple numeric that is not preceded with an alphanumeric character
# the numeric can be a float (so a dot is found) but
# it does not recognise notation such as 123e5, 0xFF, infinity or NaN
CP_NUMERIC_RE = re.compile(r'(?<![a-zA-Z0-9_"\'])([0-9\.]+)')
CP_NUMERIC_PLACEHOLDER = '__PyJsNUM_%i_PyJsNUM__'
CP_NUMERIC_PLACEHOLDER_REVERSE_RE = re.compile(
    CP_NUMERIC_PLACEHOLDER.replace('%i', r'([0-9\.]+)'))

# the re below is how we'll recognise string constants
# it finds a ' or ", then reads until the next matching ' or "
# this re only services simple cases, it can not be used when
# there are escaped quotes in the expression

#CP_STRING_1 = re.compile(r'(["\'])(.*?)\1') # this is how we'll recognise string constants

CP_STRING = '"([^\\\\"]+|\\\\([bfnrtv\'"\\\\]|[0-3]?[0-7]{1,2}|x[0-9a-fA-F]{2}|u[0-9a-fA-F]{4}))*"|\'([^\\\\\']+|\\\\([bfnrtv\'"\\\\]|[0-3]?[0-7]{1,2}|x[0-9a-fA-F]{2}|u[0-9a-fA-F]{4}))*\''
CP_STRING_RE = re.compile(
    CP_STRING)  # this is how we'll recognise string constants
CP_STRING_PLACEHOLDER = '__PyJsSTR_%i_PyJsSTR__'
CP_STRING_PLACEHOLDER_REVERSE_RE = re.compile(
    CP_STRING_PLACEHOLDER.replace('%i', r'([0-9\.]+)'))

cache = {}

# This crap is still needed but I removed it for speed reasons. Have to think ofa  better idea
# import js2py.pyjs, sys
# # Redefine builtin objects... Do you have a better idea?
# for m in list(sys.modules):
# 	if m.startswith('js2py'):
# 		del sys.modules[m]
# del js2py.pyjs
# del js2py

DEFAULT_HEADER = u'''from js2py.pyjs import *
# setting scope
var = Scope( JS_BUILTINS )
set_global_object(var)

# Code follows:
'''


def dbg(x):
    """does nothing, legacy dummy function"""
    return ''

# Another way of doing that would be with my auto esprima translation but its much slower:
# parsed = esprima.parse(js).to_dict()
def pyjsparser_parse_fn(code):
    parser = pyjsparser.PyJsParser()
    return parser.parse(code)

def translate_js(js, HEADER=DEFAULT_HEADER, use_compilation_plan=False, parse_fn=pyjsparser_parse_fn):
    """js has to be a javascript source code.
       returns equivalent python code."""
    if use_compilation_plan and not '//' in js and not '/*' in js:
        return translate_js_with_compilation_plan(js, HEADER=HEADER)

    parsed = parse_fn(js)
    translating_nodes.clean_stacks()
    return HEADER + translating_nodes.trans(
        parsed)  # syntax tree to python code


class match_unumerator(object):
    """This class ise used """
    matchcount = -1

    def __init__(self, placeholder_mask):
        self.placeholder_mask = placeholder_mask
        self.matches = []

    def __call__(self, match):
        self.matchcount += 1
        self.matches.append(match.group(0))
        return self.placeholder_mask % self.matchcount

    def __repr__(self):
        return '\n'.join(self.placeholder_mask % counter + '=' + match
                         for counter, match in enumerate(self.matches))

    def wrap_up(self, output):
        for counter, value in enumerate(self.matches):
            output = output.replace(
                "u'" + self.placeholder_mask % (counter) + "'", value, 1)
        return output


def get_compilation_plan(js):
    match_increaser_str = match_unumerator(CP_STRING_PLACEHOLDER)
    compilation_plan = re.sub(CP_STRING, match_increaser_str, js)

    match_increaser_num = match_unumerator(CP_NUMERIC_PLACEHOLDER)
    compilation_plan = re.sub(CP_NUMERIC_RE, match_increaser_num,
                              compilation_plan)
    # now put quotes, note that just patching string replaces is somewhat faster than
    # using another re:
    compilation_plan = compilation_plan.replace(
        '__PyJsNUM_', '"__PyJsNUM_').replace('_PyJsNUM__', '_PyJsNUM__"')
    compilation_plan = compilation_plan.replace(
        '__PyJsSTR_', '"__PyJsSTR_').replace('_PyJsSTR__', '_PyJsSTR__"')

    return match_increaser_str, match_increaser_num, compilation_plan


def translate_js_with_compilation_plan(js, HEADER=DEFAULT_HEADER):
    """js has to be a javascript source code.
       returns equivalent python code.

       compile plans only work with the following restrictions:
       - only enabled for oneliner expressions
       - when there are comments in the js code string substitution is disabled
       - when there nested escaped quotes string substitution is disabled, so

       cacheable:
       Q1 == 1 && name == 'harry'

       not cacheable:
       Q1 == 1 && name == 'harry' // some comment

       not cacheable:
       Q1 == 1 && name == 'o\'Reilly'

       not cacheable:
       Q1 == 1 && name /* some comment */ == 'o\'Reilly'
       """

    match_increaser_str, match_increaser_num, compilation_plan = get_compilation_plan(
        js)

    cp_hash = hashlib.md5(compilation_plan.encode('utf-8')).digest()
    try:
        python_code = cache[cp_hash]['proto_python_code']
    except:
        parser = pyjsparser.PyJsParser()
        parsed = parser.parse(compilation_plan)  # js to esprima syntax tree
        # Another way of doing that would be with my auto esprima translation but its much slower and causes import problems:
        # parsed = esprima.parse(js).to_dict()
        translating_nodes.clean_stacks()
        python_code = translating_nodes.trans(
            parsed)  # syntax tree to python code
        cache[cp_hash] = {
            'compilation_plan': compilation_plan,
            'proto_python_code': python_code,
        }

    python_code = match_increaser_str.wrap_up(python_code)
    python_code = match_increaser_num.wrap_up(python_code)

    return HEADER + python_code


def trasnlate(js, HEADER=DEFAULT_HEADER):
    """js has to be a javascript source code.
       returns equivalent python code.

       Equivalent to translate_js"""
    return translate_js(js, HEADER)


syntax_tree_translate = translating_nodes.trans

if __name__ == '__main__':
    PROFILE = False
    import js2py
    import codecs

    def main():
        with codecs.open("esprima.js", "r", "utf-8") as f:
            d = f.read()
            r = js2py.translate_js(d)

            with open('res.py', 'wb') as f2:
                f2.write(r)
            exec (r, {})

    if PROFILE:
        import cProfile
        cProfile.run('main()', sort='tottime')
    else:
        main()



from ..base import *

@Js
def console():
    pass

@Js
def log():
    print(" ".join(repr(element) for element in arguments.to_list()))

console.put('log', log)
console.put('debug', log)
console.put('info', log)
console.put('warn', log)
console.put('error', log)



class BooleanPrototype:
    def toString():
        if this.Class != 'Boolean':
            raise this.Js(TypeError)('this must be a boolean')
        return 'true' if this.value else 'false'

    def valueOf():
        if this.Class != 'Boolean':
            raise this.Js(TypeError)('this must be a boolean')
        return this.value



from __future__ import unicode_literals

from js2py.internals.conversions import *
from js2py.internals.func_utils import *


class ConsoleMethods:
    def log(this, args):
        x = ' '.join(to_string(e) for e in args)
        print(x)
        return undefined



from ..conversions import *
from ..func_utils import *


def Boolean(this, args):
    return to_boolean(get_arg(args, 0))


def BooleanConstructor(args, space):
    temp = space.NewObject()
    temp.prototype = space.BooleanPrototype
    temp.Class = 'Boolean'
    temp.value = to_boolean(get_arg(args, 0))
    return temp
